"""Create the hash table.

This program is mainly for the insert function.
"""
import base64
import hashlib
import json
import math
import os
import random
try:
    import sympy
except ImportError:
    raise ImportError("Dependency sympy is missing. Use 'pip install sympy' to install it.")
import zlib
from typing import Any, Dict, Tuple, Union

from .directories import get_full_path
from .errors import AlgorithmError, EncodingError

supported_algorithms = ["sha256", "sha512", "sha512_256", "md5"]

__all__ = [
    "insert",
    "insert_wordlists",
    "supported_algorithms"
] 

def hash_function1(hash_string, hash_dict_length) -> str:
    """Calculates a hash index."""
    seed = int.from_bytes(hash_string.encode("utf-8"), "big")
    random.seed(seed)
    hash_index = random.randint(-(2 << 32), (2 << 32)) % hash_dict_length
    return str(hash_index)

def hash_function2(hash_string, hash_dict_length) -> str:
    """Calculates a different hash index to hash_function1."""
    seed = int.from_bytes(hash_string.encode("utf-8"), "big")
    random.seed(seed)
    hash_index = random.randint(-(2 << 64), (2 << 64)) % hash_dict_length
    return str(hash_index)

def secondary_insert(collision_stash, display_progress=True, collision_dict_length_multiplier=1) -> Union[Tuple[Dict[str, str], int], int]: 
    """Insert words from the collision stash into a secondary hash table.
    
    The actual length of the dict includes the probe limit, which makes 
    it so that you don't have to do bounds checking, the table will just rehash.
    """
    collision_dict = {}
    collision_dict_length = sympy.nextprime((len(collision_stash) * 2) * collision_dict_length_multiplier)
    probe_limit = int(math.log(collision_dict_length, 2))

    for x in map(str, range(0, collision_dict_length + probe_limit)):
        collision_dict[x] = ""

    for x in map(str, range(0, len(collision_stash))):
        if display_progress == True:
            print(f"Current word: {int(x)+1}/{len(collision_stash)}")

        current_data = collision_stash[int(x)]
        current_index = hash_function1(current_data[0:current_data.find(":")], collision_dict_length)
        current_psl = 0

        while True:

            if current_psl == probe_limit:
                """Returns the collision_dict_length_multiplier so that the main function 
                can increase it and then recall this function."""
                return collision_dict_length_multiplier
            
            if collision_dict[current_index] == "":
                collision_dict[current_index] = current_data + f":{current_psl}"
                break

            else:
                collision_data = collision_dict[current_index][0:collision_dict[current_index].rfind(":")]
                collision_psl = int(collision_dict[current_index][collision_dict[current_index].rfind(":")+1:len(collision_dict[current_index])])

            if current_psl > collision_psl:
                collision_dict[current_index] = current_data + f":{current_psl}"

                current_data = collision_data
                current_psl = collision_psl + 1
                current_index = str(int(current_index) + 1)

            elif current_psl <= collision_psl:
                current_psl = current_psl + 1
                current_index = str(int(current_index) + 1)

    return (collision_dict, collision_dict_length)

def insert(wordlist, hash_type, wordlist_encoding="utf-8", display_progress=True, compression=False) -> None:
    """Insert each word in the wordlist with its corresponding hash."""
    if hash_type not in supported_algorithms:
        raise AlgorithmError(f"Hash algorithm {hash_type} is not currently supported.")

    try:
        wordlist_file = open(wordlist, "r", encoding=wordlist_encoding)
        wordlist = wordlist_file.read().splitlines()
        wordlist_file.close()
    except UnicodeDecodeError:
        raise EncodingError(f"Wordlist {wordlist} is unable to be decoded with the encoding {wordlist_encoding}.")
    except FileNotFoundError:
        raise FileNotFoundError(f"Wordlist {wordlist} not found. It needs to be in the current directory.")

    old_wordlist_length = len(wordlist)
    wordlist = list(set(wordlist)) #remove any duplicate words
    new_wordlist_length = len(wordlist)
    duplicates_found = old_wordlist_length - new_wordlist_length

    hash_file_directory = get_full_path()
    if os.path.isfile(hash_file_directory) == False:
        raise FileNotFoundError("Create the directory (create_directory) and the file (create_file) before inserting.")
    
    hash_file = open(hash_file_directory, "r")
    hash_file_content = json.loads(hash_file.read())
    hash_file.close()

    max_load_factor = 0.5

    collision_stash = []
    if hash_file_content != [{}, {}, {}, []]:
        hash_dict_length = len(hash_file_content[0])
        hash_dict1 = hash_file_content[0]
        hash_dict2 = hash_file_content[1]
        collision_dict = hash_file_content[2]
        collision_dict_length = hash_file_content[3][0]

        """Remove any words from the wordlist that are already in any of the hash dictionaries."""
        hash_dict1_values = hash_dict1.values()
        hash_dict2_values = hash_dict2.values()
        collision_dict_values = collision_dict.values()

        for word in wordlist:
            for data in hash_dict1_values:
                if data[data.find(":")+1:len(data)] == word:
                    wordlist.remove(word)
                    duplicates_found = duplicates_found + 1

            for data in hash_dict2_values:
                if data[data.find(":")+1:len(data)] == word:
                    wordlist.remove(word)
                    duplicates_found = duplicates_found + 1

            for data in collision_dict_values:
                if data[data.find(":")+1:data.rfind(":")] == word:
                    wordlist.remove(word)
                    duplicates_found = duplicates_found + 1
        
        """Calculate if we are going to go over a load factor of 0.5 when we insert the wordlist. If so, rehash before we insert."""
        values_taken = 0
        for x in map(str, range(0, hash_dict_length)):
            if hash_dict1[x] != "":
                values_taken = values_taken + 1
            
            if hash_dict2[x] != "":
                values_taken = values_taken + 1
 
        values_taken = values_taken + len(wordlist)
        load_factor = values_taken / (hash_dict_length * 2) #the load factor includes both tables

        if load_factor >= max_load_factor:
            """Rehash."""
            rehash_list = []
            for x in map(str, range(0, hash_dict_length)):
                if hash_dict1[x] != "":
                    word = hash_dict1[x][hash_dict1[x].find(":")+1:len(hash_dict1[x])]
                    rehash_list.append(word)
                
                if hash_dict2[x] != "":
                    word = hash_dict2[x][hash_dict2[x].find(":")+1:len(hash_dict2[x])]
                    rehash_list.append(word)
            rehash_list = rehash_list + wordlist

            wordlist = rehash_list
            hash_dict_length = sympy.nextprime(len(wordlist) * 2)
            hash_dict1 = {}
            hash_dict2 = {}

            for x in map(str, range(0, hash_dict_length)):
                hash_dict1[x] = ""
                hash_dict2[x] = ""
    
    else:
        hash_dict_length = sympy.nextprime(len(wordlist) * 2)
        hash_dict1 = {}
        hash_dict2 = {}
        collision_dict = {}
        collision_dict_length = 0

        for x in map(str, range(0, hash_dict_length)):
            hash_dict1[x] = ""
            hash_dict2[x] = ""

    if display_progress == True:
        print(f"Removed {duplicates_found} duplicate(s) from wordlist.")

    for x in map(str, range(0, len(wordlist))): 

        if display_progress == True:
            print(f"Current word: {int(x)+1}/{len(wordlist)}")
    
        current_word = wordlist[int(x)] 

        if hash_type == "sha256":
            hash_encoder = hashlib.new("sha256")
        elif hash_type == "sha512":
            hash_encoder = hashlib.new("sha512")
        elif hash_type == "sha512_256":
            hash_encoder = hashlib.new("sha512_256")
        elif hash_type == "md5":
            hash_encoder = hashlib.new("md5")
        
        hash_encoder.update(current_word.encode("utf-8"))
        current_hash = hash_encoder.hexdigest()
        
        data_to_insert = current_hash + ":" + current_word
        data_index1 = hash_function1(current_hash, hash_dict_length)

        if hash_dict1[data_index1] == "":
            hash_dict1[data_index1] = data_to_insert
        else:
            current_dict = "hash_dict1"
            collision = hash_dict1[data_index1]
            collision_hash = collision[0:collision.find(":")]
            collision_loop = 0
            hash_dict1[data_index1] = data_to_insert

            """Start a loop where the collisions will try and find a permanent place. 
            If the loop exceeds 10 runs, break and add to the collision stash,
            which will be sorted after the main insertion."""
            while True:

                if current_dict == "hash_dict1":
                    collision_index2 = hash_function2(collision_hash, hash_dict_length)
                    if hash_dict2[collision_index2] == "":
                        hash_dict2[collision_index2] = collision
                        break
                    else:
                        new_collision = hash_dict2[collision_index2]
                        hash_dict2[collision_index2] = collision

                        collision = new_collision
                        collision_hash = collision[0:collision.find(":")]
                        current_dict = "hash_dict2"

                elif current_dict == "hash_dict2":
                    collision_index1 = hash_function1(collision_hash, hash_dict_length)
                    if hash_dict1[collision_index1] == "":
                        hash_dict1[collision_index1] = collision
                        break
                    else:
                        new_collision = hash_dict1[collision_index1]
                        hash_dict1[collision_index1] = collision

                        collision = new_collision
                        collision_hash = collision[0:collision.find(":")]
                        current_dict = "hash_dict1"

                collision_loop = collision_loop + 1
                if collision_loop == 10:
                    collision_stash.append(collision)
                    break
   
    hash_file = open(hash_file_directory, "w")
    content_to_write = json.dumps([hash_dict1, hash_dict2, {}, [0]])
    hash_file.write(content_to_write)
    hash_file.close()
    del hash_dict1, hash_dict2 #clear some memory
   
    if collision_stash != []:
        """Insert any collision cycles into the secondary hash table."""
        if collision_dict != {}:
            for x in map(str, range(0, len(collision_dict))):
                if collision_dict[x] != "":
                    word = collision_dict[x][collision_dict[x].find(":")+1:collision_dict[x].rfind(":")]
                    collision_stash.append(word)

        response = secondary_insert(collision_stash, display_progress=display_progress)
        if isinstance(response, int):
            """Keep calling the function again and again to rehash until we dont need to anymore."""
            while True:
                collision_dict_length_multiplier = response
                response = secondary_insert(collision_stash, display_progress=display_progress, collision_dict_length_multiplier=(collision_dict_length_multiplier * 2))
                if isinstance(response, tuple):
                    """We dont need to rehash anymore."""
                    collision_dict = response[0]
                    collision_dict_length = response[1]
                    break
        else:
            collision_dict = response[0]
            collision_dict_length = response[1]
    
    hash_file = open(hash_file_directory, "r") 
    hash_file_content = json.loads(hash_file.read())
    hash_file.close()
    
    content_to_write = json.dumps([hash_file_content[0], hash_file_content[1], collision_dict, [collision_dict_length]])
    if compression == True:
        """Compress the content so it takes up less space in the file."""
        content_to_write = content_to_write.encode("utf-8")
        content_to_write = zlib.compress(content_to_write, 9)
        content_to_write = base64.b64encode(content_to_write)
        content_to_write = content_to_write.decode("utf-8")

    hash_file = open(hash_file_directory, "w")
    hash_file.write(content_to_write)
    hash_file.close()

def insert_wordlists(wordlists, hash_type, wordlist_encoding="utf-8", display_progress=True, compression=False) -> None:
    """Insert a list of wordlists."""
    if not isinstance(wordlists, list):
        raise TypeError(f"Wordlists input must be type list, not type {type(wordlists).__name__}.")

    for wordlist in wordlists:
        insert(wordlist, hash_type, wordlist_encoding=wordlist_encoding, display_progress=display_progress, compression=compression)


"""Hashtable with worst case O(log n) lookup.

This hashtable works similarly to a rainbow table, where it stores plaintext values
with the corresponding hash, meaning you can 'reverse' the hash.

The main table uses the Cuckoo implementation, and any cycles from 
that table go to the secondary table, which uses the Robin Hood implementation.

Go to the README.md for usage.
"""
from .directories import *
from .errors import *
from .insert import *
from .search import *


"""Manage the directories

This includes options to store the hashtable in a file or keep 
it temporarily, where it will be returned from the function...
"""
import json
import os
import platform
from typing import Tuple, Union

from .errors import FilenameError, PathError, SystemNotSupported

__all__ = [
    "set_directory",
    "set_filename",
    "get_directory",
    "get_filename",
    "get_full_path",
    "create_directory",
    "create_file"
]

directory = os.getcwd()
filename = "hash_table.json"

def check_for_invalid_characters(string, os, path_input=False) -> Tuple[bool, str]:
    """Give an error if any illegal characters are detected based on the os.
    
    Forward and back slashes will be ignored since they are used to define each section in a path.
    """
    if os == "Windows":
        if path_input == True:
            illegal_characters = ["<", ">", ":", '"', "|", "?", "*"]
        elif path_input == False:
            illegal_characters = ["<", ">", ":", '"', "/", "\\", "|", "?", "*"]

    elif os == "Linux":
        if path_input == False:
            illegal_characters = ["/"]
        else:
            return (False, "")

    for character in illegal_characters:
        if character in string:
            if character == ":":
                for x in range(0, len(string)):
                    if string[x] == ":" and x != 1:
                        return (True, ":")
            else:
                return (True, character)
    return (False, "")

def check_for_reserved_names(string, os) -> Tuple[bool, str]:
    """Give an error if any reserved names are used in the file path."""
    if os == "Windows":
        illegal_names = [
            "CON", "PRN", "AUX", "NUL", "COM1", "COM2", "COM3", 
            "COM4", "COM5", "COM6", "COM7", "COM8", "COM9", "LPT1",
            "LPT2", "LPT3", "LPT4", "LPT5", "LPT6", "LPT7", "LPT8", "LPT9"
        ]
    elif os == "Linux":
        """Linux has no reserved names."""
        return (False, "")
    else:
        raise SystemNotSupported(f"Unsupported OS {os}.")

    """An exception should only be given if an illegal name is found on its own with no other characters."""
    for reserved_name in illegal_names:
        if string.upper() == reserved_name:
            return (True, reserved_name)

        if reserved_name in string.upper():
            reserved_name_index = string.upper().find(reserved_name)

            if string[reserved_name_index-1] == "\\" or string[reserved_name_index-1] == "/":
                if reserved_name_index+(len(reserved_name)-1) == len(string)-1:
                    return (True, reserved_name)

                if string[reserved_name_index+len(reserved_name)] == "\\" or string[reserved_name_index+len(reserved_name)] == "/":
                    return (True, reserved_name)

    return (False, "")

def set_directory(file_directory, full_path=False) -> None:
    """Sets the directory of where the file will be stored. If 'full_path' is true,
    set the directory as the full path.

    If no directory is set, the file will be stored in the program location.
    """
    global directory

    invalid_characters = check_for_invalid_characters(file_directory, platform.system(), path_input=True)
    if invalid_characters[0] == True:
        raise PathError(f"Invalid character ({invalid_characters[1]}) found in file path.")

    reserved_names = check_for_reserved_names(file_directory, platform.system())
    if reserved_names[0] == True:
        raise PathError(f"OS reserved name ({reserved_names[1]}) found in file path.")

    if full_path == False:
        directory = directory + file_directory

    elif full_path == True:
        directory = file_directory

def set_filename(file_name) -> None:
    """Sets the filename.
    
    If no filename is set, it will remain the default name..
    """
    global filename

    file_name = file_name.strip()
    if file_name == "":
        raise FilenameError("The filename cannot be set as an empty value.")

    invalid_characters = check_for_invalid_characters(file_name, platform.system())
    if invalid_characters[0] == True:
        raise FilenameError(f"Invalid character ({invalid_characters[1]}) found in filename.")

    reserved_names = check_for_reserved_names(file_name, platform.system())
    if reserved_names[0] == True:
        raise FilenameError(f"OS reserved name ({reserved_names[1]}) found in filename.")
    
    filename = file_name + ".json"

def get_directory() -> str:
    """Returns the current set directory."""
    return directory

def get_filename(file_extension=False) -> str:
    """Returns the current set filename."""
    if file_extension == False:
        return filename[0:filename.find(".json")]
    else:
        return filename

def get_full_path(file_extension=True) -> str:
    """Returns the current set directory and filename."""
    return get_directory() + "/" + get_filename(file_extension=file_extension)

def create_directory() -> None:
    """Creates the set/default directory."""
    slash_indexes = []
    for x in range(0, len(directory)):
        if directory[x] == "/" or directory[x] == "\\":
            slash_indexes.append(x)
    
    directories_to_create = []
    for x in range(0, len(slash_indexes)):
        if x == len(slash_indexes)-1:
            if os.path.isdir(directory[0:len(directory)]):
                existing_directory = directory[0:len(directory)]
            else:
                directories_to_create.append(directory[0:len(directory)])

        else:     
            if os.path.isdir(directory[0:slash_indexes[x+1]]):
                existing_directory = directory[0:slash_indexes[x+1]]
            else:
                directories_to_create.append(directory[0:slash_indexes[x+1]])

    for _dir in directories_to_create:
        os.mkdir(_dir)

def create_file(overwrite_existing=False) -> Union[bool, None]:
    """Creates the file inside of the set/default directory.
    
    Only proceed if the set/default directory exists.
    """
    if os.path.isfile(get_full_path()):
        if overwrite_existing == False:
            return False

    if os.path.isdir(directory):
        file_path = get_full_path()
        create_file = open(file_path, "w")
        create_file.write(json.dumps([{}, {}, {}, []]))
        create_file.close()
    
    else:
        raise PathError("The directory must be created (.create_directory()) before the file is.")


"""Search the hash table.

This program is mainly for the search function.
"""
import base64
import json
import math
import os
import time
import zlib
from typing import Any, Tuple, Union

from .errors import SearchError
from .insert import hash_function1, hash_function2

__all__ = [
    "search"
]

def search(hash_string, hash_table_path, full_path=False, time_took=True, compression=False) -> Union[Tuple[str, str], str, bool]: #time_tooks is measured in seconds
    """Search through the hash table until we find a plaintext match to the hash given."""
    start_time = time.time()
    
    hash_table_path = hash_table_path + ".json"
    if full_path == False:
        hash_table_path = os.getcwd() + "/" + hash_table_path

    if os.path.isfile(hash_table_path) == False:
        raise FileNotFoundError("Create the directory (create_directory) and the file (create_file) before inserting.")

    hash_file = open(hash_table_path, "r")
    hash_file_content = hash_file.read()
    if compression == True:
        """Decompress the hash file content."""
        hash_file_content = hash_file_content.encode("utf-8")
        hash_file_content = base64.b64decode(hash_file_content)
        hash_file_content = zlib.decompress(hash_file_content)
        hash_file_content = hash_file_content.decode("utf-8")
    hash_file_content = json.loads(hash_file_content)
    hash_file.close()

    if hash_file_content == [{}, {}, {}, []]:
        raise SearchError("Insert data into the hash table before doing a lookup.")

    hash_dict1 = hash_file_content[0]
    hash_dict2 = hash_file_content[1]
    collision_dict = hash_file_content[2]
    hash_dict_length = len(hash_dict1)
    collision_dict_length = hash_file_content[3][0]

    hash_index1 = hash_function1(hash_string, hash_dict_length) 
    hash_index2 = hash_function2(hash_string, hash_dict_length)
    if collision_dict_length > 0:
        probe_limit = int(math.log(collision_dict_length, 2))
        hash_index3 = hash_function1(hash_string, collision_dict_length) #for the secondary table

    if hash_dict1 != {}:
        if hash_string == hash_dict1[hash_index1][0:hash_dict1[hash_index1].find(":")]:
            matching_plaintext = hash_dict1[hash_index1][(hash_dict1[hash_index1].find(":"))+1:len(hash_dict1[hash_index1])]
            end_time = time.time()

            if time_took == True:
                return (matching_plaintext, str(end_time-start_time))
            else:
                return matching_plaintext

    if hash_dict2 != {}:
        if hash_string == hash_dict2[hash_index2][0:hash_dict2[hash_index2].find(":")]:
            matching_plaintext = hash_dict2[hash_index2][(hash_dict2[hash_index2].find(":"))+1:len(hash_dict2[hash_index2])]
            end_time = time.time()

            if time_took == True:
                return (matching_plaintext, str(end_time-start_time))
            else:
                return matching_plaintext

    if collision_dict_length > 0:
        for x in map(str, range(int(hash_index3), int(hash_index3)+probe_limit+1)):
            if hash_string == collision_dict[x][0:collision_dict[x].find(":")]:
                matching_plaintext = collision_dict[x][collision_dict[x].find(":")+1:collision_dict[x].rfind(":")]
                end_time = time.time()

                if time_took == True:
                    return (matching_plaintext, str(end_time-start_time))
                else:
                    return matching_plaintext

    return False


from setuptools import setup

setup(
    name="rainbowtables",
    version="1.0.6",
    author="JustBen",
    author_email="justben009@gmail.com",
    description="A python library allowing the user to create a rainbowtable.",
    keywords="rainbowtable",
    python_requires=">=3.7",
    packages=["rainbowtables"],
    long_description=open("README.md", "r", encoding="utf-8").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/JustBennnn/rainbowtables",
    project_urls={
        "Issue Tracker": "https://github.com/JustBennnn/rainbowtables/issues",
    },
    install_requires=["sympy"],
    license="MIT",
    classifiers=[
        "Programming Language :: Python :: 3",
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Natural Language :: English",
        "Topic :: Software Development :: Libraries",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Utilities",
    ],
)



"""Hashtable errors.

These are custom error messages.
"""
class HashtableException(Exception):
    """Base class for other custom exceptions."""
    pass

class FilenameError(HashtableException):
    """An invalid character, reserved name or an empty input in the filename."""
    pass

class PathError(HashtableException):
    """An invalid character or reserved name found in the path."""
    pass

class EncodingError(HashtableException):
    """The wordlist is unable to be decoded with the encoding entered."""
    pass

class AlgorithmError(HashtableException):
    """The algorithm they have tried to use is not supported."""
    pass

class SearchError(HashtableException):
    """Any errors given from the search function."""

class SystemNotSupported(HashtableException):
    """Unable to detect specific values for the current OS."""